ClassDescription subclass:#JavaClass
	instanceVariableNames:'fullName accessFlags constantPool sourceFile fields initialized
		initValues staticFields'
	classVariableNames:'InitialValuePerType'
	poolDictionaries:''
	category:'Java-Classes'
!


!JavaClass  class methodsFor:'initialization'!

initialize
    InitialValuePerType := IdentityDictionary new.
    InitialValuePerType at:$B put:0.
    InitialValuePerType at:$C put:0.
    InitialValuePerType at:$D put:0.0.
    InitialValuePerType at:$F put:0.0 asShortFloat.
    InitialValuePerType at:$I put:0.
    InitialValuePerType at:$J put:0.
    InitialValuePerType at:$S put:0.
    InitialValuePerType at:$Z put:0.
    InitialValuePerType at:$L put:nil.
    InitialValuePerType at:$[ put:nil.

    "
     JavaClass initialize
    "
! !

!JavaClass  class methodsFor:'instance creation'!

fullName:aString
    ^ self new setFullName:aString

    "Created: 15.4.1996 / 15:52:55 / cg"
!

fullName:aString numStatic:nStatic
    |meta cls|

    "create the metaclass first"

    meta := Metaclass new.
    meta setSuperclass:(self).
    meta instSize:(JavaClass "class" instSize + nStatic).
    meta setName:(aString , 'class') asSymbol.
    meta setClassVariableString:''.

    "then let the new meta create the class"
    cls := meta new.
    cls setSuperclass:JavaObject.
    cls instSize:0.
    cls setName:aString.
    cls flags:0.
"/    cls setClassVariableString:''.
    cls setInstanceVariableString:''.
    cls category:#java.

    cls setFullName:aString.
    ^ cls

"/    ^ self new setFullName:aString

    "Created: 15.4.1996 / 15:52:55 / cg"
!

name:aString
    self shouldNotImplement

    "Created: 15.4.1996 / 15:52:55 / cg"
! !

!JavaClass  class methodsFor:'signature parsing'!

initialValueFromSignature:aSignature
    "given a signature, return an initializer value"

    |s|

    s := aSignature readStream.
    ^ self initialValueFromStream:s.

    "
     JavaClass initialValueFromSignature:'LObject;'    
     JavaClass initialValueFromSignature:'B'        
     JavaClass initialValueFromSignature:'I'        
    "


!

initialValueFromStream:s
    "parse a fieldTypeSpec - see java doc"

    |typeChar|

    typeChar := s next.
    ^ InitialValuePerType at:typeChar ifAbsent:nil.
! !

!JavaClass  class methodsFor:'special'!

setInstanceVariableStringFromFields:f in:aClass
    |varNames|

    varNames := ''.
    f do:[:aField |
        varNames := varNames , aField name , ' '
    ].
    aClass setInstanceVariableString:varNames.

    "Created: 15.4.1996 / 16:42:52 / cg"
! !

!JavaClass methodsFor:'accessing'!

category
    "java classes do not really hav a category;
     simulate one by separating its name into individual components,
     prepending 'java'"

    |nameComponents|

    nameComponents := fullName asCollectionOfSubstringsSeparatedBy:$/.
    ^ 'java-' , (nameComponents asStringWith:$/ from:1 to:(nameComponents size - 1))
!

constantPool
    ^ constantPool
!

fullName
    ^ fullName
!

name
    ^ name "/ fullName
!

package
    "extract from the fullName"

    |components|

    components := fullName asCollectionOfSubstringsSeparatedBy:$/.
    components size > 1 ifTrue:[
        ^ (components copyWithoutLast:1) asStringWith:$/
    ].
    ^ fullName

    "
     Java allClasses first fullName
     Java allClasses first package  
    "
!

sourceFile
    ^ sourceFile
! !

!JavaClass methodsFor:'java initialization'!

classInit
    "call the classes JAVA clinit function"

    |m|

    initialized := true.

    m := self compiledMethodAt:#'<clinit>()V'.
    m notNil ifTrue:[
        'calling clinit() of ' print. self fullName printNL.
        [
            self invokeJavaMethod:m.
        ] valueOnUnwindDo:[
            initialized := false
        ]
    ] ifFalse:[
        self fullName print. ' has no clinit()' printNL.
    ].

    "
     JavaInterpreter instructionTrace:true.
     JavaInterpreter callTrace:true.

     (Java classNamed:'java.lang.String') classInit
     (Java classNamed:'java.lang.System') classInit

     (Java classNamed:'java/io/FileDescriptor') classInit
     (Java classNamed:'java.util.Properties') classInit 
    "
!

initializeStaticFields
    staticFields do:[:f |
        f constantValue notNil ifTrue:[
            self instVarNamed:(f name) put:f constantValue.
        ]
    ].
! !

!JavaClass methodsFor:'java instance creation'!

initValueFor:instVarName
    |idx field|

    idx := fields findFirst:[:field | field name = instVarName].
    idx == 0 ifTrue:[
        superclass ~~ JavaObject ifTrue:[
            ^ superclass initValueFor:instVarName
        ].
        self halt
    ] ifFalse:[
        field := fields at:idx.
        ^ field initialValue
    ].
!

new
    "create a new instance, and call its JAVA init function"

    |newJavaObject|

    newJavaObject := self newCleared.
    newJavaObject invoke:#'<init>'.
    ^ newJavaObject

    "
     (Java classNamed:'java.lang.String') basicNew inspect
     (Java classNamed:'java.lang.String') new inspect
    "


!

newCleared
    "create a new cleared JAVA instance"

    |newJavaObject|

    newJavaObject := super basicNew.

    fields isNil ifTrue:[
        'OOPS - no fieldSpec for new object' errorPrintNL.
        newJavaObject initializeToZero.     "/ mhmh
    ] ifFalse:[
        newJavaObject initializeFields:initValues
    ].
    ^ newJavaObject

    "
     (Java classNamed:'java.lang.String') basicNew inspect
     (Java classNamed:'java.lang.String') newCleared inspect
     (Java classNamed:'java.lang.String') new inspect
    "
! !

!JavaClass methodsFor:'message sending'!

invokeJava:selector
    "send javaSelector (name+sig) message, without arguments"

    |method|

    method := self lookupMethodFor:selector.
    method notNil ifTrue:[
        ^ self invokeJavaMethod:method
    ].

    ^ self doesNotUnderstand:(Message selector:selector)


!

invokeJava:selector with:arg
    "send javaSelector (name+sig) message, with 1 argument"

    |method|

    method := self lookupMethodFor:selector.
    method notNil ifTrue:[
        ^ self invokeJavaMethod:method with:arg
    ].
    ^ self doesNotUnderstand:(Message selector:selector)

!

invokeJavaMethod:aJavaMethod
    "invoke a static java method, without arguments"

    |i val|

    aJavaMethod numArgs ~~ 0 ifTrue:[
        self halt:'need arguments'
    ].
    aJavaMethod isStatic ifFalse:[
        self halt:'non-static function'
    ].

    i := JavaInterpreter new.
    val := i interpret:aJavaMethod sender:thisContext sender.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)


!

invokeJavaMethod:aJavaMethod with:arg
    "invoke a static java method, with one argument"

    |i val|

    aJavaMethod numArgs ~~ 1 ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifFalse:[
        self halt:'non-static function'
    ].

    i := JavaInterpreter new.
    i push:arg.

    val := i interpret:aJavaMethod sender:thisContext sender.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)


!

lookupMethodFor:selector
    |method cls sel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        cls := self.
        [cls notNil] whileTrue:[
            method := cls compiledMethodAt:sel.
            method notNil ifTrue:[ ^ method ].
            cls := cls superclass.
        ].
    ].

    ^ nil
! !

!JavaClass methodsFor:'printing & storing'!

displayString
    ^ fullName copy replaceAll:$/ by:$.
"/    ^ 'JAVA-' , name .
"/    ^ name , '(Java)'  "/ 'JavaClass(' , name , ')'

    "Created: 15.4.1996 / 16:02:48 / cg"
! !

!JavaClass methodsFor:'private accessing'!

addMethod:m name:name signature:signature
    |sel|

    sel := (name , signature) asSymbol.
    self addSelector:sel withMethod:m.
    ^ self.

    selectorArray isNil ifTrue:[
        selectorArray := #().
        methodArray := #()
    ].
    selectorArray := selectorArray copyWith:(name , signature) asSymbol. "/ name.
"/    selectorArray := selectorArray copyWith:name asSymbol. "/ name.
    methodArray := methodArray copyWith:m

    "Created: 15.4.1996 / 16:42:52 / cg"
!

markUninitialized
    initialized := false.
!

setAccessFlags:flags
    accessFlags := flags.

    "Created: 15.4.1996 / 16:42:52 / cg"
!

setConstantPool:anArray
    constantPool := anArray.

    "Created: 15.4.1996 / 16:42:52 / cg"
!

setFields:f
    JavaClass setInstanceVariableStringFromFields:f in:self.

    instSize := superclass instSize + f size.

    fields := f.
    initValues := self allInstVarNames collect:[:nm |
        self initValueFor:nm
    ].

    "Created: 15.4.1996 / 16:42:52 / cg"


!

setFullName:aString
    |nameComponents|

    fullName := aString.
    nameComponents := aString asCollectionOfSubstringsSeparatedBy:$/.
    name := nameComponents last

    "Created: 15.4.1996 / 16:42:52 / cg"
!

setSourceFile:aFilename
    sourceFile := aFilename.

    "Created: 15.4.1996 / 16:42:52 / cg"
!

setStaticFields:f
    JavaClass setInstanceVariableStringFromFields:f in:self class.

    staticFields := f.


!

setSuperclass:aClass
    aClass isNil ifTrue:[
        super setSuperclass:JavaObject
    ] ifFalse:[
        (aClass isMemberOf:JavaUnresolvedClassConstant) ifTrue:[
            self halt.
        ].
        super setSuperclass:aClass.
    ].

    instvars notNil ifTrue:[
        (superclass notNil and:[(superclass isMemberOf:JavaUnresolvedClassConstant) not]) ifTrue:[
            instSize := superclass instSize + (instvars asCollectionOfWords size)
        ] ifFalse:[
            instSize := instvars asCollectionOfWords size
        ]
    ].

    "Created: 15.4.1996 / 16:42:52 / cg"
! !

!JavaClass methodsFor:'queries'!

browserClass
    "return the browser to use for this class - 
     this can be redefined in special classes, to get different browsers"

    ^ JavaBrowser




!

hasInterface:aJavaInterface
    aJavaInterface selectorArray do:[:sel |
        (self implements:sel) ifFalse:[
            self halt.
            ^ false.
        ]
    ].
    ^ true
!

isClass
    ^ true
!

isFinal
    ^ (accessFlags bitAnd:16r0010) ~~ 0
!

isInitialized
    ^ initialized == true
!

isInterface
    ^ (accessFlags bitAnd:16r0200) ~~ 0
!

isPublic
    ^ (accessFlags bitAnd:16r0001) ~~ 0
! !

!JavaClass methodsFor:'resolving class refs'!

resolveClassRefsIgnoring:classesBeingLoaded
    |newClass const newConst|

    (superclass isMemberOf:JavaUnresolvedClassConstant) ifTrue:[
        newClass := superclass resolve.
        newClass isNil ifTrue:[
            newClass := JavaClassReader 
                                loadClassLazy:(superclass fullName)
                                ignoring:classesBeingLoaded.
        ].
        newClass notNil ifTrue:[
            superclass := newClass
        ]
    ].

    1 to:constantPool size do:[:index |
        const := constantPool at:index.
        (const isMemberOf:JavaUnresolvedClassConstant) ifTrue:[
            "/ try to resolve
            newConst := const resolve.
            newConst isNil ifTrue:[
                (classesBeingLoaded includes:(const fullName)) ifFalse:[
                    newConst := JavaClassReader 
                                    loadClassLazy:(const fullName)
                                    ignoring:classesBeingLoaded.
                    ]
            ].
            newConst notNil ifTrue:[
                constantPool at:index put:newConst
            ]
        ]
    ].

    (superclass notNil and:[(superclass isMemberOf:JavaUnresolvedClassConstant) not]) ifTrue:[
        superclass resolveClassRefsIgnoring:classesBeingLoaded.
        instSize := superclass instSize + (instvars asCollectionOfWords size)
    ] ifFalse:[
        instSize := instvars asCollectionOfWords size
    ].
"/Transcript show:'resolved '; show: self displayString; 
"/           show: ' instSize='; showCr:  instSize.
! !

!JavaClass methodsFor:'smalltalk interface'!

invoke:selector
    "send a message, without args"

    |method cls sel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        cls := self.
        [cls notNil and:[cls ~~ JavaObject]] whileTrue:[
            cls methodArray do:[:aMethod |
"/     aMethod name printNL.
                aMethod name == selector ifTrue:[
                    aMethod numArgs == 0 ifTrue:[
                        ^ self invokeJavaMethod:aMethod
                    ]
                ]
            ].
            cls := cls superclass.
        ].
    ].

    ^ self doesNotUnderstand:(Message selector:selector)

    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#'<init>'. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#isEmpty. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#size. 
    "


! !

!JavaClass  class methodsFor:'documentation'!

version
    ^ '$Header: /home/jv/Projects/SmalltalkX/repositories/cvs/stx/libjava/JavaClass.st,v 1.22 1996/06/24 15:10:57 cg Exp $'
! !
JavaClass initialize!
